# Copyright (C) 2025  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import logging

import click

# WARNING: do not import unnecessary things here to keep cli startup time under
# control


logger = logging.getLogger(__name__)

# marker of a deleted/non-populated index entry
NULLKEY = b"\x00" * 32

CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])

try:
    # make this cli usable both from the swh.core's 'swh' cli group and from
    # direct swh-shard command (since swh-shard does not depend on swh.core)
    from swh.core.cli import swh

    cli_group = swh.group
except (ImportError, ModuleNotFoundError):
    cli_group = click.group


@cli_group(name="shard", context_settings=CONTEXT_SETTINGS)
@click.pass_context
def shard_cli_group(ctx):
    """Software Heritage Shard tools."""


@shard_cli_group.command("info")
@click.argument(
    "shard", required=True, nargs=-1, type=click.Path(exists=True, dir_okay=False)
)
@click.pass_context
def shard_info(ctx, shard):
    "Display shard file information"

    from swh.shard import Shard

    for shardfile in shard:
        with Shard(shardfile) as s:
            h = s.header
            click.echo(f"Shard {shardfile}")
            click.echo(f"├─version:    {h.version}")
            click.echo(f"├─objects:    {h.objects_count}")
            click.echo(f"│ ├─position: {h.objects_position}")
            click.echo(f"│ └─size:     {h.objects_size}")
            click.echo("├─index")
            click.echo(f"│ ├─position: {h.index_position}")
            click.echo(f"│ └─size:     {h.index_size}")
            click.echo("├─hash")
            click.echo(f"│ └─position: {h.hash_position}")
            click.echo(f"└─end:        {s.endpos}")


@shard_cli_group.command("truncate")
@click.argument(
    "shard", required=True, nargs=-1, type=click.Path(exists=True, dir_okay=False)
)
@click.option(
    "--assume-yes",
    "--yes",
    "-y",
    is_flag=True,
    default=False,
    help=(
        "Do not ask for confirmation before actually truncating "
        "shard files (default is to ask)"
    ),
)
@click.pass_context
def shard_truncate(ctx, shard, assume_yes):
    "Truncate shard file to its minimal size, if needed"

    import os

    from swh.shard import Shard

    for shardfile in shard:
        with Shard(shardfile) as s:
            realsize = s.endpos + 1
        fsize = os.stat(shardfile).st_size
        if fsize > realsize:
            click.echo(
                f"Shard file {shardfile} is {fsize-realsize} bytes bigger than necessary"
            )
            if assume_yes or click.confirm(
                click.style("Truncate?", fg="yellow", bold=True)
            ):
                try:
                    with open(shardfile, "r+b") as fobj:
                        fobj.seek(realsize)
                        fobj.truncate()
                    click.echo(f"Truncated. New size is {realsize}")
                except OSError:
                    click.echo("Could not truncate the file. Check file permissions.")
            else:
                click.echo("Skipped")
        else:
            click.echo(
                f"Shard file {shardfile} does not seem to be overallocated, nothing to do"
            )


@shard_cli_group.command("create")
@click.argument(
    "shard", required=True, type=click.Path(exists=False, dir_okay=False, writable=True)
)
@click.argument("files", metavar="files", required=True, nargs=-1)
@click.option(
    "--sorted/--no-sorted",
    "sort_files",
    default=False,
    help=(
        "Sort files by inversed filename before adding them to the shard; "
        "it may help having better compression ratio when compressing "
        "the shard file"
    ),
)
@click.pass_context
def shard_create(ctx, shard, files, sort_files):
    "Create a shard file from given files"

    import hashlib
    import os
    import sys

    from swh.shard import ShardCreator

    if os.path.exists(shard):
        raise click.ClickException(f"Shard file {shard} already exists. Aborted!")

    files = list(files)
    if files == ["-"]:
        # read file names from stdin
        files = [fname.strip() for fname in sys.stdin.read().splitlines()]
    click.echo(f"There are {len(files)} entries")
    hashes = set()
    files_to_add = {}
    with click.progressbar(files, label="Checking files to add") as bfiles:
        for fname in bfiles:
            try:
                with open(fname, "rb") as f:
                    sha256 = hashlib.sha256(f.read()).digest()
                    if sha256 not in hashes:
                        files_to_add[fname] = sha256
                        hashes.add(sha256)
            except OSError:
                continue
    click.echo(f"after deduplication: {len(files_to_add)} entries")

    with ShardCreator(shard, len(files_to_add)) as shard:
        it = files_to_add.items()
        if sort_files:
            it = sorted(it, key=lambda x: x[0][-1::-1])
        with click.progressbar(it, label="Adding files to the shard") as items:
            for fname, sha256 in items:
                with open(fname, "rb") as f:
                    shard.write(sha256, f.read())
    click.echo("Done")


@shard_cli_group.command("ls")
@click.option("--skip-removed", default=False, is_flag=True)
@click.argument("shard", required=True, type=click.Path(exists=True, dir_okay=False))
@click.pass_context
def shard_list(ctx, skip_removed, shard):
    "List objects in a shard file"

    from swh.shard import Shard

    with Shard(shard) as s:
        for key in s:
            if skip_removed and key == NULLKEY:
                continue
            try:
                size = s.getsize(key)
            except KeyError:
                size = "N/A"
            click.echo(f"{key.hex()}: {size} bytes")


@shard_cli_group.command("get")
@click.argument("shard", required=True, type=click.Path(exists=True, dir_okay=False))
@click.argument("keys", required=True, nargs=-1)
@click.pass_context
def shard_get(ctx, shard, keys):
    "List objects in a shard file"

    from swh.shard import Shard

    with Shard(shard) as s:
        for key in keys:
            click.echo(s[bytes.fromhex(key)], nl=False)


@shard_cli_group.command("delete")
@click.argument(
    "shard", required=True, type=click.Path(exists=True, dir_okay=False, writable=True)
)
@click.argument("keys", required=True, nargs=-1)
@click.option(
    "--confirm/--no-confirm",
    default=True,
    help="Ask for confirmation before performing the deletion",
)
@click.pass_context
def shard_delete(ctx, shard, keys, confirm):
    """Delete objects from a shard file

    Keys to delete from the shard file are expected to be given as hex
    representation. If there is only one argument '-', then read the list of
    keys from stdin. Implies --no-confirm.

    If at least one key is missing or invalid, the whole process is aborted.

    """
    import sys

    if keys == ("-",):
        keys = sys.stdin.read().split()
        confirm = False
    if len(set(keys)) < len(keys):
        ctx.fail("There are duplicate keys, aborting")

    from swh.shard import Shard

    obj_size = {}
    with Shard(shard) as s:
        for key in keys:
            try:
                obj_size[key] = s.getsize(bytes.fromhex(key))
            except ValueError:
                click.secho(f"{key}: key is invalid", fg="red")
            except KeyError:
                click.secho(f"{key}: key not found", fg="red")
    if len(obj_size) < len(keys):
        raise click.ClickException(
            "There have been errors for at least one key, aborting"
        )
    click.echo(f"About to remove these objects from the shard file {shard}")
    for key in keys:
        click.echo(f"{key} ({obj_size[key]} bytes)")
    if confirm:
        click.confirm(
            click.style(
                "Proceed?",
                fg="yellow",
                bold=True,
            ),
            abort=True,
        )
    with click.progressbar(keys, label="Deleting objects from the shard") as barkeys:
        for key in barkeys:
            Shard.delete(shard, bytes.fromhex(key))
    click.echo("Done")


@shard_cli_group.command("check")
@click.argument(
    "shard", required=True, nargs=-1, type=click.Path(exists=True, dir_okay=False)
)
@click.option(
    "--with-hash", is_flag=True, default=False, help="Also check objects hashes (slow!)"
)
@click.pass_context
def shard_check(ctx, shard, with_hash):
    "Check a shard file for data corruption or inconsistencies"

    import hashlib
    import struct

    from swh.shard import Shard

    errors = []
    sizes = []
    for shardfile in shard:
        with Shard(shardfile) as s:
            h = s.header
            n_entries = h.index_size // 40  # 32 + 8
            idx = h.index_position
            obj_pos = h.objects_position
            logger.debug("IDX=%s", idx)
            logger.debug("OBJS=%s", obj_pos)
            logger.debug("ENTRIES=%s", n_entries)

            # for most of the checks we only use direct access to the file,
            # without using the C extension _shard, but for checking the
            # hashmap; so we have the shard file open twice here: one as a
            # regular python (binary) file object, and one via the Shard object
            # encapsulating the C extension.
            with open(shardfile, "rb") as fob:
                # build the list of object offsets as built in the *objects*
                # section of the shard file
                obj_offsets = []
                fob.seek(h.objects_position)
                with click.progressbar(
                    length=h.objects_size, label="listing objects"
                ) as bar:
                    while True:
                        obj_offsets.append(fob.tell())
                        size = struct.unpack(">Q", fob.read(8))[0]
                        sizes.append(size)
                        fob.seek(size, 1)
                        bar.update(8 + size)
                        if (fob.tell() - h.objects_position) >= h.objects_size:
                            break
                # build the list of object offsets as defined in the *Index*
                # section of the shard file. If asked for, also check each object's
                # sha256 sum while iterating on the index entries.
                idx_offsets = []
                fob.seek(h.index_position)
                with click.progressbar(
                    length=h.index_size, label="listing indexes"
                ) as bar:
                    while True:
                        key = fob.read(32)
                        offset = fob.read(8)
                        if key != NULLKEY:
                            idx_offsets.append(struct.unpack(">Q", offset)[0])
                            pos = fob.tell()
                            fob.seek(idx_offsets[-1])
                            size = struct.unpack(">Q", fob.read(8))[0]
                            try:
                                # calling Shard.getsize() will use the hashmap,
                                # thus checking it's OK for the key
                                s.getsize(key)
                            except Exception as exc:
                                errors.append(
                                    f"{key.hex()}: Shard.getsize() error ({exc})"
                                )
                            if with_hash:
                                content_hash = hashlib.sha256(fob.read(size)).digest()
                                if content_hash != key:
                                    errors.append(
                                        f"{key.hex()}: hash mismatch (computed: {content_hash.hex()})"  # noqa
                                    )
                            fob.seek(pos)
                        # next position (key (32) + offset (8))
                        bar.update(40)
                        if (fob.tell() - h.index_position) >= h.index_size:
                            break
        # compare the 2 offset lists; once sorted, they should match exactly
        if len(idx_offsets) != len(obj_offsets):
            errors.append(
                f"Offset lists len mismatch: {len(idx_offsets)} vs. {len(obj_offsets)}"
            )
        if sorted(idx_offsets) != obj_offsets:
            errors.append("Offset lists mismatch")

    if (sum(sizes) + h.objects_count * 8) != h.objects_size:
        errors.append(
            f"Total size mismatch: {sum(sizes)} + ({h.objects_count} * 8) computed "
            f"vs. {h.objects_size} advertised"
        )
    if errors:
        click.echo(
            "check failed with the following error(s):\n"
            + "\n".join(f"- {err}" for err in errors),
            err=True,
        )
        ctx.exit(1)


def main():
    # Even though swh() sets up logging, we need an earlier basic logging setup
    # for the next few logging statements
    logging.basicConfig()
    return shard_cli_group(auto_envvar_prefix="SWH")


if __name__ == "__main__":
    main()



# Copyright (C) 2025 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from hashlib import sha256
import struct

from click.testing import CliRunner
import pytest

from swh.shard import Shard, ShardCreator, cli


@pytest.fixture
def small_shard(tmp_path):
    with ShardCreator(str(tmp_path / "small.shard"), 16) as shard:
        for i in range(16):
            shard.write(bytes.fromhex(f"{i:-064X}"), bytes((65 + i,)) * 42)
    # add a bit of extra padding, to check shard end pos computation is OK
    with (tmp_path / "small.shard").open("ab") as f:
        f.write(b"\x00" * 100)
    return tmp_path / "small.shard"


@pytest.fixture
def small_shard_ro(small_shard):
    small_shard.chmod(0o444)
    try:
        yield small_shard
    finally:
        small_shard.chmod(0o644)


@pytest.fixture
def valid_shard(tmp_path):
    with ShardCreator(str(tmp_path / "valid.shard"), 16) as shard:
        for i in range(16):
            data = bytes((65 + i,)) * 11 * (i + 1)
            key = sha256(data).digest()
            shard.write(key, data)
    return tmp_path / "valid.shard"


@pytest.fixture
def corrupted_shard_hash(tmp_path):
    """create a corrupted shard file

    It's coming with:
    - at index 5: a corrupted data (hash of the content does not match the key)
    """
    fname = tmp_path / "corrupted_hash.shard"
    objid = 5
    with ShardCreator(str(fname), 16) as shard:
        for i in range(16):
            data = bytes((65 + i,)) * 11 * (i + 1)
            key = sha256(data).digest()
            if i == objid:
                # bitflip: same size, only one bit altered
                data = data[:-1] + (data[-1] ^ 1).to_bytes(length=1, byteorder="big")
                altered_key = key
            shard.write(key, data)

    return fname, objid, altered_key


@pytest.fixture
def corrupted_shard_size(tmp_path):
    """create a corrupted shard file

    It's coming with:
    - at index 7: an invalid object size
    """
    fname = tmp_path / "corrupted_size.shard"
    objid = 7
    with ShardCreator(str(fname), 16) as shard:
        for i in range(16):
            data = bytes((65 + i,)) * 11 * (i + 1)
            key = sha256(data).digest()
            shard.write(key, data)
            if i == objid:
                corr_key = key
    # change the size of object at index 7
    with Shard(str(fname)) as s:
        idx = s.getindex(s.getpos(corr_key))
        assert corr_key == idx.key
        offset = idx.object_offset
    with open(fname, "r+b") as shardfile:
        shardfile.seek(offset)
        size = struct.unpack(">Q", shardfile.read(8))[0]
        newsize = struct.pack(">Q", size + 1)
        shardfile.seek(-8, 1)
        shardfile.write(newsize)

    return fname, objid, corr_key


@pytest.fixture
def corrupted_shard_offset(tmp_path):
    """create a corrupted shard file

    It's coming with:
    - at index 3: an invalid offset
    """
    fname = tmp_path / "corrupted_offset.shard"
    objid = 3
    keys = []
    with ShardCreator(str(fname), 16) as shard:
        for i in range(16):
            data = bytes((65 + i,)) * 11 * (i + 1)
            key = sha256(data).digest()
            shard.write(key, data)
            keys.append(key)

    # change the offset of object at index 3. For this, use the offset of
    # another existing object (here obj5) so the file is not completely
    # garbage...
    with Shard(str(fname)) as s:
        corr_key = keys[objid]
        obj_pos = s.getpos(corr_key)
        use_offset = s.getindex(s.getpos(keys[(objid + 2) % 16])).object_offset
        idx_pos = s.header.index_position

    with open(fname, "r+b") as shardfile:
        # set the fp at beginning of the offset for object number 3
        shardfile.seek(idx_pos + 40 * obj_pos + 32)
        shardfile.write(struct.pack(">Q", use_offset))

    return fname, objid, corr_key


def test_cli():
    runner = CliRunner()
    result = runner.invoke(cli.shard_cli_group)
    assert result.exit_code == 2
    assert "Software Heritage Shard tools" in result.output


def test_cli_info(small_shard):
    runner = CliRunner()
    result = runner.invoke(cli.shard_info, [str(small_shard)])
    assert result.exit_code == 0
    assert (
        result.output
        == f"""\
Shard {small_shard}
├─version:    1
├─objects:    16
│ ├─position: 512
│ └─size:     800
├─index
│ ├─position: 1312
│ └─size:     680
├─hash
│ └─position: 1992
└─end:        2070
"""
    )


def test_cli_truncate(small_shard):
    runner = CliRunner()
    # first time, not validating the operation (hitting return, means saying N)
    # note: explicitly adding the \n is required for the test to pass on
    # jenkins (because of no tty or something like that)
    result = runner.invoke(cli.shard_truncate, [str(small_shard)], input="\n")
    assert result.exit_code == 0, result.output
    assert (
        result.output
        == f"""\
Shard file {small_shard} is 100 bytes bigger than necessary
Truncate? [y/N]: \nSkipped
"""
    )
    # Using explicit \n above to defeat auto-trim feature of code editors...

    # second time to shrink the file
    result = runner.invoke(cli.shard_truncate, [str(small_shard)], input="y")
    assert result.exit_code == 0, result.output
    assert (
        result.output
        == f"""\
Shard file {small_shard} is 100 bytes bigger than necessary
Truncate? [y/N]: y
Truncated. New size is 2071
"""
    )

    # second time is a noop
    result = runner.invoke(cli.shard_truncate, [str(small_shard)], input="y")
    assert result.exit_code == 0, result.output
    assert (
        result.output
        == f"""\
Shard file {small_shard} does not seem to be overallocated, nothing to do
"""
    )


@pytest.mark.parametrize("option", ["--assume-yes", "--yes", "-y"])
def test_cli_truncate_assume_yes(small_shard, option):
    runner = CliRunner()
    result = runner.invoke(cli.shard_truncate, [option, str(small_shard)])
    assert result.exit_code == 0, result.output
    assert (
        result.output
        == f"""\
Shard file {small_shard} is 100 bytes bigger than necessary
Truncated. New size is 2071
"""
    )


def test_cli_truncate_perm_error(small_shard_ro):
    runner = CliRunner()
    result = runner.invoke(cli.shard_truncate, ["-y", str(small_shard_ro)])
    assert result.exit_code == 0, result.output
    assert (
        result.output
        == f"""\
Shard file {small_shard_ro} is 100 bytes bigger than necessary
Could not truncate the file. Check file permissions.
"""
    )


def test_cli_ls(small_shard):
    runner = CliRunner()
    result = runner.invoke(cli.shard_list, [str(small_shard)])
    assert result.exit_code == 0
    assert set(result.output.strip().splitlines()) == {
        "0000000000000000000000000000000000000000000000000000000000000000: 42 bytes",
        "0000000000000000000000000000000000000000000000000000000000000001: 42 bytes",
        "0000000000000000000000000000000000000000000000000000000000000002: 42 bytes",
        "0000000000000000000000000000000000000000000000000000000000000003: 42 bytes",
        "0000000000000000000000000000000000000000000000000000000000000004: 42 bytes",
        "0000000000000000000000000000000000000000000000000000000000000005: 42 bytes",
        "0000000000000000000000000000000000000000000000000000000000000006: 42 bytes",
        "0000000000000000000000000000000000000000000000000000000000000007: 42 bytes",
        "0000000000000000000000000000000000000000000000000000000000000008: 42 bytes",
        "0000000000000000000000000000000000000000000000000000000000000009: 42 bytes",
        "000000000000000000000000000000000000000000000000000000000000000a: 42 bytes",
        "000000000000000000000000000000000000000000000000000000000000000b: 42 bytes",
        "000000000000000000000000000000000000000000000000000000000000000c: 42 bytes",
        "000000000000000000000000000000000000000000000000000000000000000d: 42 bytes",
        "000000000000000000000000000000000000000000000000000000000000000e: 42 bytes",
        "000000000000000000000000000000000000000000000000000000000000000f: 42 bytes",
    }


def test_cli_get(small_shard):
    runner = CliRunner()
    for i in range(16):
        result = runner.invoke(cli.shard_get, [str(small_shard), f"{i:-064x}"])
        assert result.exit_code == 0
        assert result.output == chr(65 + i) * 42


def test_cli_create(tmp_path):
    runner = CliRunner()

    files = []
    hashes = []
    for i in range(16):
        f = tmp_path / f"file_{i}"
        data = f"file {i}".encode()
        f.write_bytes(data)
        files.append(str(f))
        hashes.append(sha256(data).digest())
    shard = tmp_path / "shard"
    result = runner.invoke(cli.shard_create, [str(shard), *files])
    assert result.exit_code == 0
    assert result.output.strip().endswith("Done")
    with Shard(str(shard)) as s:
        assert s.header.objects_count == 16
        # check stored sha256 digests are as expected
        assert sorted(list(s)) == sorted(hashes)


def test_cli_delete_one_abort(small_shard):
    runner = CliRunner()
    key_num = 5
    key = f"{key_num:-064X}"
    result = runner.invoke(
        cli.shard_delete,
        [str(small_shard), key],
        input="n\n",
    )
    assert result.exit_code == 1, result.output
    assert "Proceed? [y/N]" in result.output
    assert "Aborted!" in result.output

    result = runner.invoke(cli.shard_get, [str(small_shard), key])
    assert result.exit_code == 0
    assert result.output == chr(65 + key_num) * 42


def test_cli_delete_invalid_key_abort(small_shard):
    runner = CliRunner()
    keys = [f"{i:-064x}" for i in range(5)]
    keys.append("00" * 16)
    result = runner.invoke(
        cli.shard_delete,
        [str(small_shard), *keys],
    )
    assert result.exit_code == 1, result.output
    assert "key is invalid" in result.output
    assert "aborting" in result.output


def test_cli_delete_unknown_key_abort(small_shard):
    runner = CliRunner()
    keys = [f"{i:-064x}" for i in range(5)]
    keys.append("01" * 32)
    result = runner.invoke(
        cli.shard_delete,
        [str(small_shard), *keys],
    )
    assert result.exit_code == 1, result.output
    assert "key not found" in result.output
    assert "aborting" in result.output


@pytest.mark.parametrize("key_nums", [(5,), (1, 3, 5), tuple(range(16))])
def test_cli_delete_confirm(small_shard, key_nums):
    runner = CliRunner()
    keys = [f"{key_num:-064x}" for key_num in key_nums]
    result = runner.invoke(
        cli.shard_delete,
        [str(small_shard), *keys],
        input="y\n",
    )
    assert result.exit_code == 0, result.output
    assert "Proceed? [y/N]" in result.output
    assert "Done" in result.output

    result = runner.invoke(cli.shard_list, [str(small_shard)])
    assert result.exit_code == 0
    for i in range(16):
        key = f"{i:-064x}"
        if i in key_nums:
            assert key not in result.output
        else:
            assert key in result.output


@pytest.mark.parametrize("key_nums", [(5,), (1, 3, 5), tuple(range(16))])
def test_cli_delete_from_stdin(small_shard, key_nums):
    runner = CliRunner()
    keys = [f"{key_num:-064x}" for key_num in key_nums]
    result = runner.invoke(
        cli.shard_delete,
        [str(small_shard), "-"],
        input="\n".join(keys),
    )
    assert result.exit_code == 0, result.output
    assert "Proceed? [y/N]" not in result.output
    assert "Done" in result.output

    result = runner.invoke(cli.shard_list, [str(small_shard)])
    assert result.exit_code == 0
    for i in range(16):
        key = f"{i:-064x}"
        if i in key_nums:
            assert key not in result.output
        else:
            assert key in result.output


def test_cli_delete_one_no_confirm(small_shard):
    runner = CliRunner()
    key_num = 5
    key = f"{key_num:-064x}"
    result = runner.invoke(
        cli.shard_delete,
        ["--no-confirm", str(small_shard), key],
    )
    assert result.exit_code == 0, result.output
    assert "Proceed? [y/N]" not in result.output
    assert "Done" in result.output

    result = runner.invoke(cli.shard_list, [str(small_shard)])
    assert result.exit_code == 0
    for i in range(16):
        key = f"{i:-064x}"
        if i == key_num:
            assert key not in result.output
        else:
            assert key in result.output


@pytest.mark.parametrize("with_hash", [False, True])
def test_cli_check_ok(valid_shard, with_hash):
    runner = CliRunner()
    args = []
    if with_hash:
        args.append("--with-hash")
    args.append(str(valid_shard))
    result = runner.invoke(cli.shard_check, args)
    assert result.exit_code == 0, result.output


def test_cli_check_corrupted_ok_no_hash(corrupted_shard_hash):
    shardfile, objid, key = corrupted_shard_hash
    runner = CliRunner()
    args = [str(shardfile)]
    result = runner.invoke(cli.shard_check, args)
    assert result.exit_code == 0, result.output


def test_cli_check_corrupted_ko_hash(corrupted_shard_hash):
    shardfile, objid, key = corrupted_shard_hash
    runner = CliRunner()
    args = ["--with-hash", str(shardfile)]
    result = runner.invoke(cli.shard_check, args)
    assert result.exit_code != 0, result.output
    assert f"{key.hex()}: hash mismatch" in result.output


def test_cli_check_corrupted_ko_size(corrupted_shard_size):
    shardfile, objid, key = corrupted_shard_size
    runner = CliRunner()
    args = [str(shardfile)]
    result = runner.invoke(cli.shard_check, args)
    assert result.exit_code != 0, result.output
    assert "Total size mismatch" in result.output
    assert "Offset lists mismatch" in result.output


def test_cli_check_corrupted_ko_offset(corrupted_shard_offset):
    shardfile, objid, key = corrupted_shard_offset
    runner = CliRunner()
    args = ["--with-hash", str(shardfile)]
    result = runner.invoke(cli.shard_check, args)
    assert result.exit_code != 0, result.output
    assert "Offset lists mismatch" in result.output
    # object with invalid offset in the index should generate an invalid hash
    # error as well
    assert f"{key.hex()}: hash mismatch" in result.output



def pytest_addoption(parser):
    parser.addoption(
        "--shard-size",
        default=10,
        type=int,
        help="Size of the Read Shard file in MB",
    )
    parser.addoption(
        "--shard-path",
        default="/tmp/payload",
        help="Path of the Read Shard file",
    )
    parser.addoption(
        "--shard-count",
        default=2,
        type=int,
        help="Number of Read Shard files for lookup tests",
    )
    parser.addoption(
        "--object-max-size",
        default=10 * 1024,
        type=int,
        help="Maximum size of an object in bytes",
    )
    parser.addoption(
        "--lookups",
        default=10,
        type=int,
        help="Number of lookups to perform",
    )



# Copyright (C) 2021-2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import gc
from hashlib import sha256
import logging
import os
from pathlib import Path
import platform
import random
import resource
import time

import pytest

from swh.shard import Shard, ShardCreator

logger = logging.getLogger(__name__)


PYPY = platform.python_implementation() == "PyPy"


@pytest.fixture(scope="function", autouse=True)
def setrlimit(request):
    marker = request.node.get_closest_marker("setrlimit")
    rlimits = []
    if marker is not None:
        for which, (soft, hard) in marker.args:
            backup = resource.getrlimit(which)
            logger.info("Saving rlimit %s (%s, %s)", which, *backup)
            rlimits.append((which, backup))
            logger.info("Setting rlimit %s (%s, %s)", which, soft, hard)
            resource.setrlimit(which, (soft, hard))

    yield

    for which, (soft, hard) in rlimits:
        logger.info("Resetting rlimit %s (%s, %s)", which, soft, hard)
        resource.setrlimit(which, (soft, hard))
        result = resource.getrlimit(which)
        logger.info("Resulting rlimit %s (%s, %s)", which, *result)


KEY_A = b"A" * Shard.key_len
KEY_B = b"B" * Shard.key_len
KEY_C = b"C" * Shard.key_len

OBJECT_A = b"AAAA"
OBJECT_B = b"BBBB"
# Use a size bigger than BUF_SIZE used in shard_write_zeros()
OBJECT_C = b"C" * 10_000


@pytest.fixture
def populated_shard_path(tmp_path):
    shard_path = tmp_path / "shard"
    shard_path.open("w").close()
    os.truncate(shard_path, 10 * 1024 * 1024)

    with ShardCreator(str(shard_path), 3) as s:
        s.write(KEY_A, OBJECT_A)
        s.write(KEY_B, OBJECT_B)
        s.write(KEY_C, OBJECT_C)

    return str(shard_path)


def test_lookup(populated_shard_path):
    with Shard(populated_shard_path) as s:
        assert s.lookup(KEY_A) == OBJECT_A
        assert s.lookup(KEY_B) == OBJECT_B
        assert s.lookup(KEY_C) == OBJECT_C


def test_creator_open_without_permission(tmpdir):
    path = Path(tmpdir / "no-perm")
    path.touch()
    # Remove all permissions
    path.chmod(0o000)
    exc_cls = RuntimeError if PYPY else PermissionError
    with pytest.raises(exc_cls):
        with ShardCreator(str(path), 1):
            pass


@pytest.mark.setrlimit((resource.RLIMIT_FSIZE, (64_000, -1)))
def test_write_above_rlimit_fsize(tmpdir):
    with pytest.raises(OSError, match=r"File too large"):
        with ShardCreator(f"{tmpdir}/test-shard", 1) as shard:
            shard.write(b"A" * Shard.key_len, b"A" * 72_000)


def test_write_errors_if_too_many(tmpdir):
    with ShardCreator(f"{tmpdir}/shard", 1) as shard:
        shard.write(b"A" * Shard.key_len, b"AAAA")
        with pytest.raises(ValueError):
            shard.write(b"B" * Shard.key_len, b"BBBB")


def test_write_errors_if_not_enought(tmpdir):
    with pytest.raises(RuntimeError):
        with ShardCreator(f"{tmpdir}/shard", 2) as shard:
            shard.write(b"A" * Shard.key_len, b"AAAA")


def test_write_errors_for_wrong_key_len(tmpdir):
    with ShardCreator(f"{tmpdir}/shard", 1) as shard:
        with pytest.raises(ValueError):
            shard.write(b"A", b"AAAA")
        # add a valid entry so the __exit__ does not raise an exception
        shard.write(b"A" * Shard.key_len, b"AAAA")


def test_creator_context_does_not_run_finalize_on_error(tmpdir, mocker):
    pytest.skip("Not sure how to do that...")
    import contextlib

    mock_method = mocker.patch.object(ShardCreator, "__exit__")
    with contextlib.suppress(KeyError):
        with ShardCreator(f"{tmpdir}/shard", 1) as _:
            raise KeyError(42)
    mock_method.assert_not_called()


@pytest.mark.setrlimit((resource.RLIMIT_FSIZE, (64_000, -1)))
def test_finalize_above_rlimit_fsize(tmpdir):
    path = f"{tmpdir}/shard"
    with pytest.raises(OSError, match="File too large"):
        with ShardCreator(path, 1) as shard:
            shard.write(b"A" * Shard.key_len, b"A" * 63_500)


def test_creator_errors_with_duplicate_key(tmpdir):
    with pytest.raises(RuntimeError, match="duplicate"):
        with ShardCreator(f"{tmpdir}/shard", 2) as shard:
            shard.write(b"A" * Shard.key_len, b"AAAA")
            shard.write(b"A" * Shard.key_len, b"BBBB")


def test_load_non_existing():
    exc_cls = RuntimeError if PYPY else FileNotFoundError
    with pytest.raises(exc_cls):
        _ = Shard("/nonexistent")


@pytest.fixture
def corrupted_shard_path(tmpdir):
    # taken from hash.h
    SHARD_OFFSET_HEADER = 512
    path = f"{tmpdir}/corrupted"
    with ShardCreator(path, 1) as s:
        s.write(b"A" * Shard.key_len, b"AAAA")
    with open(path, "rb+") as f:
        f.seek(SHARD_OFFSET_HEADER)
        # replace the object size (uint64_t) by something larger than file size
        f.write(b"\x00\x00\x00\x00\x00\x00\xff\xff")
    return path


def test_lookup_failure(corrupted_shard_path):
    with Shard(corrupted_shard_path) as shard:
        with pytest.raises(RuntimeError, match=r"failed.*corrupted"):
            shard.lookup(b"A" * Shard.key_len)


def test_lookup_errors_for_wrong_key_len(tmpdir):
    with pytest.raises(ValueError):
        with ShardCreator(f"{tmpdir}/shard", 1) as shard:
            shard.write(b"A", b"AAAA")


@pytest.fixture
def shard_with_mismatched_key(tmp_path):
    path = tmp_path / "mismatched"
    with ShardCreator(str(path), 1) as s:
        s.write(b"A" * Shard.key_len, b"AAAA")
    # Replace the key in the index
    content = path.read_bytes()
    path.write_bytes(content.replace(b"A" * Shard.key_len, b"B" * Shard.key_len))
    return str(path)


def test_lookup_errors_for_mismatched_key(shard_with_mismatched_key):
    with Shard(shard_with_mismatched_key) as shard:
        with pytest.raises(KeyError):
            shard.lookup(b"A" * Shard.key_len)


@pytest.fixture
def shard_with_deleted_objects_path(populated_shard_path):
    Shard.delete(populated_shard_path, KEY_A)
    Shard.delete(populated_shard_path, KEY_C)
    return populated_shard_path


def test_delete_has_zeroed_object_data(shard_with_deleted_objects_path):
    shard_content = open(shard_with_deleted_objects_path, "rb").read()
    assert b"AAAA" not in shard_content
    assert b"CCCCCCCC" not in shard_content


def test_delete_leaves_other_keys_working(shard_with_deleted_objects_path):
    with Shard(shard_with_deleted_objects_path) as shard:
        assert shard.lookup(KEY_B) == OBJECT_B


def test_lookup_on_deleted_raises_key_error(shard_with_deleted_objects_path):
    with Shard(shard_with_deleted_objects_path) as shard:
        with pytest.raises(KeyError):
            _ = shard.lookup(KEY_A)
        with pytest.raises(KeyError):
            _ = shard.lookup(KEY_C)


@pytest.fixture
def payload(request):
    size = request.config.getoption("--shard-size")
    path = request.config.getoption("--shard-path")
    if not os.path.exists(path) or os.path.getsize(path) != size * 1024 * 1024:
        os.system(f"dd if=/dev/urandom of={path} count={size} bs=1024k")
    return path


#
# PYTHONMALLOC=malloc valgrind --tool=memcheck .tox/py3/bin/pytest \
#    -k test_build_speed swh/perfecthash/tests/test_hash.py |& tee /tmp/v
#
def test_build_speed(request, tmpdir, payload):
    start = time.time()
    os.system(f"cp {payload} {tmpdir}/shard")
    baseline = time.time() - start
    write_duration, build_duration, _ = shard_build(request, tmpdir, payload)
    duration = write_duration + build_duration
    print(
        f"baseline {baseline}, "
        f"write_duration {write_duration}, "
        f"build_duration {build_duration}, "
        f"total_duration {duration}"
    )
    #
    # According to the docs/benchmarks.rst analysis, the duration is below 5
    # times the baseline time. This assertion is here to ensure we do not not
    # regress in the future... (we use x10 to give a bit of slack otherwise the
    # test is pretty unstable)
    #
    k = 100 if PYPY else 10
    assert duration < baseline * k


def test_lookup_speed(request, tmpdir, payload):
    _, _, objects = shard_build(request, tmpdir, payload)
    for i in range(request.config.getoption("--shard-count")):
        os.system(f"cp {tmpdir}/shard {tmpdir}/shard{i}")
    start = time.time()
    shard_lookups(request, tmpdir, objects)
    duration = time.time() - start

    lookups = request.config.getoption("--lookups")
    key_per_sec = lookups / duration
    print(f"key lookups speed = {key_per_sec:.2f}/s")


def shard_lookups(request, tmpdir, objects):
    shard_path = f"{tmpdir}/shard"
    shards = []
    for i in range(request.config.getoption("--shard-count")):
        shards.append(Shard(f"{shard_path}{i}"))
    lookups = request.config.getoption("--lookups")
    count = 0
    while True:
        for key, object_size in objects.items():
            if count >= lookups:
                return
            for shard in shards:
                object = shard.lookup(key)
                assert len(object) == object_size
                count += 1
    for shard in shards:
        shard.close()


def shard_build(request, tmpdir, payload):
    shard_size = request.config.getoption("--shard-size") * 1024 * 1024
    shard_path = f"{tmpdir}/shard"
    open(shard_path, "w").close()
    os.truncate(shard_path, shard_size * 2)

    object_max_size = request.config.getoption("--object-max-size")
    objects = {}
    count = 0
    size = 0
    keys = []
    with open(payload, "rb") as f:
        while True:
            object = f.read(random.randrange(512, object_max_size))
            if len(object) < 512:
                break
            key = sha256(object).digest()
            objects[key] = len(object)
            size += len(object)
            count += 1
            keys.append(key)

    print(f"number of objects = {count}, total size = {size}")
    assert size <= shard_size
    start = time.time()

    with ShardCreator(shard_path, len(objects)) as shard:
        count = 0
        size = 0
        with open(payload, "rb") as f:
            for key in keys:
                object = f.read(objects[key])
                assert len(object) == objects[key]
                count += 1
                size += len(object)
                shard.write(key, object)
        write_duration = time.time() - start
        start = time.time()

    build_duration = time.time() - start
    return write_duration, build_duration, objects


def test_memleak(request, tmpdir, payload):
    "Naive test for memleak in ShardReader"
    shard_build(request, tmpdir, payload)
    shard_file = str(tmpdir / "shard")
    maxrss = [resource.getrusage(resource.RUSAGE_SELF).ru_maxrss]
    for i in range(100):
        # run the garbage collector first to avoid the memory of one iteration
        # to affect the next one
        gc.collect()
        with Shard(shard_file) as s:
            for key in s:
                obj = s[key]
                assert sha256(obj).digest() == key
        maxrss.append(resource.getrusage(resource.RUSAGE_SELF).ru_maxrss)
    # on pypy, the used rss can still grow significantly during the first gew
    # iterations, but should remain under a reasonable threshold
    assert (maxrss[-1] - maxrss[1]) < 100 * 1024  # in kB
    # but there should be none for the last iterations
    for i in range(99, 90, -1):
        assert (maxrss[i] - maxrss[i - 1]) == 0, maxrss[i] - maxrss[i - 1]



# Copyright (C) 2021-2025  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from ._shard import ShardCreator, ShardReader

__all__ = ["Shard", "ShardCreator"]


class Shard(ShardReader):
    # for BW compat reason, implement the context manager protocol
    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def __iter__(self):
        # iterate of the keys
        for i in range(self.header.index_size // (32 + 8)):  # KEY_LEN + uint64
            idx = self.getindex(i)
            if idx.object_offset < (2**64 - 1):
                yield idx.key
